<?php
namespace Studip\Dockerized\Commands;

use Studip\Dockerized\Config;
use Studip\Dockerized\Creators\DockerComposeConfiguration;
use Studip\Dockerized\Creators\NginxConfiguration;
use Studip\Dockerized\Traits\GetCompiledPath;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\ArrayInput;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;

final class Compile extends Command
{
    use GetCompiledPath;

    protected function configure()
    {
        $this->setName('compile');
        $this->setDescription('Compiles the configuration files');
        $this->addOption(
            'path', '
            p',
            InputOption::VALUE_OPTIONAL,
            'Path to store the compiled files',
            $this->getCompiledPath()
        );
        $this->addOption('nginx', null, InputOption::VALUE_NEGATABLE, 'Compile nginx configuration', true);
        $this->addOption('docker', null, InputOption::VALUE_NEGATABLE, 'Compile docker configuration', true);

        $this->addOption('build', 'b', InputOption::VALUE_NEGATABLE, 'Whether to build the containers');
        $this->addOption('force', 'F', InputOption::VALUE_NONE, 'Force compilation');
        $this->addOption('start', 's', InputOption::VALUE_NONE, 'Start container after compilation (requires "build")');
    }

    protected function initialize(InputInterface $input, OutputInterface $output)
    {
        if (!file_exists($input->getOption('path')) && !mkdir($input->getOption('path'))) {
            $output->writeln('<error>Could not create compile directory</error>');
            return Command::FAILURE;
        }

        if (!is_dir($input->getOption('path'))) {
            $output->writeln('<error>Given compile directory is invalid (not a directory)</error>');
            return Command::FAILURE;
        }

        if (!is_writable($input->getOption('path'))) {
            $output->writeln('<error>Given compile directory is invalid (not writable)</error>');
            return Command::FAILURE;
        }
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $compiledPath = realpath($input->getOption('path'));
        $force = $input->getOption('force');
        $io = new SymfonyStyle($input, $output);

        $writeConfigFile = function (string $filename, string $content) use ($compiledPath, $force, $io): bool {
            $filepath = "{$compiledPath}/{$filename}";

            if (
                !file_exists($filepath)
                || $content !== file_get_contents($filepath)
                || $force
            ) {
                file_put_contents($filepath, $content);
                $io->success('Config file ' . $filename . ' written');

                return true;
            }

            return false;
        };

        $changed = 0;
        if ($input->getOption('nginx')) {
            $changed += (int) $writeConfigFile(
                'nginx.conf',
                $this->createNginxConfiguration()
            );

            $changed += (int) $writeConfigFile(
                'nginx-sites.conf',
                $this->createNginxSites()
            );
        }

        if ($input->getOption('docker')) {
            $changed += (int) $writeConfigFile(
                'docker-compose.yml',
                $this->createDockerComposerConfiguration(
                    realpath(__DIR__ . '/../..'),
                    $compiledPath
                )
            );
        }

        if ($changed === 0) {
            $io->comment('No changes');
        }
        if ($input->getOption('build')) {
            $this->getApplication()->doRun(
                new ArrayInput(['command' => 'docker:build']),
                $output
            );

            if ($input->getOption('start')) {
                $this->getApplication()->doRun(
                    new ArrayInput(['command' => 'docker:start']),
                    $output
                );
            }
        }

        return Command::SUCCESS;
    }

    private function getPHPVersionIndex(string $version, string $prefix = ''): string
    {
        return $prefix . preg_replace('/\D/', '', $version);
    }

    public function createDockerComposerConfiguration(string $cwd, string $compiledPath): string
    {
        $creator = new DockerComposeConfiguration('studip-dockerized');

        $creator->hook('services', function (array $service): array {
            $service['restart'] = $service['restart'] ?? 'unless-stopped';
            return $service;
        });

        // Declare volumes
        $volumes = [
            [realpath($cwd . '/web'), '/usr/share/nginx/html/studip-dockerized-config.app', 'ro'],
            [realpath($cwd . '/logs'), '/usr/share/nginx/html/logs'],
            [realpath($cwd . '/config.json'), '/usr/share/nginx/html/studip-dockerized-config.json', 'ro'],
            ...array_map(
                fn(string $path, array $definition) => [
                    $definition['source'], '/usr/share/nginx/html/' . $path
                ],
                array_keys(Config::getInstance()->get('sites') ?? []),
                array_values(Config::getInstance()->get('sites') ?? [])
            )
        ];

        // Nginx
        $creator->addService('nginx', 'nginx:alpine', [
            'container_name' => 'nginx',
            'depends_on' => ['redis-server', 'memcached-server'],
            'ports' => array_merge(
                Config::getInstance()->get('default') ? ['80:80', '443:443'] : [],
                ...array_map(
                    fn($config) => [
                        "{$config[0]}:{$config[0]}",
                        "{$config[1]}:{$config[1]}"
                    ],
                    array_values(Config::getInstance()->get('php'))
                ),
            ),
        ]);

        // DB
        $dbExposingPort = (int) (Config::getInstance()->get('db')['port'] ?? 3306);

        $creator->addService('db', 'mariadb', [
            'container_name' => 'db',
            'command'        => '--sql_mode="NO_ENGINE_SUBSTITUTION"',
            'ports'          => [$dbExposingPort . ':3306'],
            'restart'        => 'unless-stopped',
            'volumes'        => ['db-data:/var/lib/mysql'],
            'environment'    => [
                'MARIADB_ROOT_PASSWORD' => 'rootpwd',
                'MARIADB_USER' => 'studip',
                'MARIADB_PASSWORD' => 'studip',
                'MARIADB_MYSQL_LOCALHOST_USER' => 'root',
                'MARIADB_AUTO_UPGRADE'  => true,
                'TZ' => 'Europe/Berlin',
            ],
            'healthcheck'    => [
                'test'         => ["CMD", "healthcheck.sh", "--connect", "--innodb_initialized"],
                'start_period' => '10s',
                'interval'     => '10s',
                'timeout'      => '10s',
                'retries'      => 3,
            ]
        ]);
        $creator->addVolume('db-data');

        $mounts = [
            "{$compiledPath}/nginx.conf" => 'conf.d/default.conf',
            "{$compiledPath}/nginx-sites.conf" => 'sites.conf',
            "{$cwd}/config/nginx-php.conf" => 'php.conf',
            "{$compiledPath}/ssl.crt" => "ssl.crt",
            "{$compiledPath}/ssl.key" => "ssl.key",
        ];
        foreach ($mounts as $source => $target) {
            if (!str_starts_with($target, '/')) {
                $target = "/etc/nginx/{$target}";
            }
            $creator->addServiceVolume('nginx', $source, $target, 'ro');
        }
        $creator->addServiceVolume('nginx', realpath("{$cwd}/logs/nginx"), '/var/log/nginx');

        foreach ($volumes as $volume) {
            $creator->addServiceVolume('nginx', ...$volume);
        }

        // PHP
        foreach (Config::getInstance()->get('php') as $version => $port) {
            $index = $this->getPHPVersionIndex($version, 'php');

            $creator->addService($index, null, [
                'container_name' => "php-{$version}",
                'build' => [
                    'context' => '.',
                    'dockerfile' => "{$cwd}/config/docker-files/{$version}-Dockerfile",
                ],
                'environment' => [
                    'MYSQL_HOST' => 'db',
                ],
                'depends_on' => [
                    'db' => [
                        'condition' => 'service_healthy',
                    ]
                ],
            ]);

            $creator->addServiceVolume(
                $index,
                "{$cwd}/config/php-fpm.conf",
                '/usr/local/etc/php-fpm.d/zz-config.conf',
                'ro'
            );
            $creator->addServiceVolume(
                $index,
                "{$cwd}/config/php.ini",
                '/usr/local/etc/php/conf.d/studip.ini',
                'ro'
            );
            $creator->addServiceVolume(
                $index,
                "{$cwd}/logs/php",
                '/var/log/'
            );

            foreach ($volumes as $volume) {
                $creator->addServiceVolume($index, ...$volume);
            }
        }

        // Other needed images
        $creator->addService('memcached-server', 'memcached:latest', [
            'container_name' => 'memcached',
        ]);

        $creator->addService('redis-server', 'redis:latest', [
            'container_name' => 'redis',
            'restart'        => 'always',
            'volumes'        => ['cache-data:/data'],
        ]);
        $creator->addVolume('cache-data');

        $creator->addService('adminer', 'adminer', [
            'container_name' => 'adminer',
            'restart'        => 'always',
            'ports'          => ['8044:8080'],
        ]);

        $creator->addService('dozzle', 'amir20/dozzle:latest', [
            'container_name' => 'dozzle',
            'volumes'        => ['/var/run/docker.sock:/var/run/docker.sock'],
            'ports'          => ['8045:8080'],
        ]);

        return $creator->dump();
    }

    public function createNginxConfiguration(): string
    {
        $creator = new NginxConfiguration();
        $creator->hook('server', function (array $config): array {
            $config['config'] = [
                ...$config['config'],

                'server_name _',

                'proxy_read_timeout 1200',
                'proxy_connect_timeout 1200',
                'proxy_send_timeout 1200',
                'send_timeout 1200',

                'access_log off',
                'error_log off',

                'include sites.conf',
            ];
            return $config;
        });

        // Add upstreams
        $servers = [];
        foreach (Config::getInstance()->get('php') as $version => [$port, $sslPort]) {
            $index = $this->getPHPVersionIndex($version, 'php');

            $creator->addUpstream("{$index}_backend", [
                "server {$index}:9000"
            ]);

            $servers[] = [$index, $port, $sslPort];

            if (Config::getInstance()->get('default') === $version) {
                $servers[] = [$index, 80, 443];
            }
        }

        // Add servers
        foreach ($servers as [$index, $port, $sslPort]) {
            $creator->addServer([
                'listen ' . $port . ' default_server',
                'client_max_body_size 64M',
                'set $fastcgi_backend ' . $index . '_backend',
            ]);

            $creator->addServer([
                'listen ' . $sslPort . ' default_server ssl',
                'http2 on',
                'client_max_body_size 64M',
                'set $fastcgi_backend ' . $index . '_backend',

                'ssl_certificate /etc/nginx/ssl.crt',
                'ssl_certificate_key /etc/nginx/ssl.key',
            ]);
        }

        return $creator->dump();
    }

    private function createNginxSites(): string
    {
        $creator = new NginxConfiguration();

        // Add default location
        $creator->addLocation('/sites', [
            'alias /usr/share/nginx/html/studip-dockerized-config.app',
            'index index.php index.html index.html',
            'set $site_document_root /usr/share/nginx/html/studip-dockerized-config.app',
            'include php.conf',
        ]);

        // Add configured locations
        foreach (Config::getInstance()->get('sites') ?? [] as $path => $definition) {
            $p = '/usr/share/nginx/html/' . $path;
            if (isset($definition['mount'])) {
                $p .= '/' . $definition['mount'];
            }

            $creator->addLocation($path, [
                'alias ' . $p,
                'index index.php index.html index.html',
                'set $site_document_root ' . $p,
                'include php.conf',
            ]);
        }

        return $creator->dump();
    }
}
