<?php
namespace Studip\Dockerized\Commands\Sites;

use Exception;
use Studip\Dockerized\Config;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Question\Question;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\Filesystem\Path;

final class Add extends Command
{
    protected function configure()
    {
        $this->setName('site:add');
        $this->setDescription('Adds a site (needs recompile)');
        $this->addArgument('webpath', InputArgument::OPTIONAL, 'The web path', '');
        $this->addArgument('localpath', InputArgument::OPTIONAL, 'The local path', '');
        $this->addArgument('mount', InputArgument::OPTIONAL, 'The mounted folder', null);
    }

    protected function execute(InputInterface $input, OutputInterface $output)
    {
        $io = new SymfonyStyle($input, $output);

        $question = $this->getWebPathQuestion();
        $webPath = $input->hasArgument('webpath')
            ? $question->getValidator()($input->getArgument('webpath'))
            : $io->askQuestion($question);

        $question = $this->getLocalPathQuestion();
        $localPath = $input->hasArgument('webpath')
            ? $question->getValidator()($input->getArgument('localpath'))
            : $io->askQuestion($question);

        $mount = $input->hasArgument('webpath')
            ? $input->getArgument('mount')
            : $io->ask('Mounted folder:');

        $this->addSiteToConfig($webPath, $localPath, $mount);

        $io->success('The site "' . $webPath . '" was added.');

        return Command::SUCCESS;
    }

    private function getWebPathQuestion(): Question
    {
        $question = new Question('Web path');
        $question->setValidator(function (string $webPath): string {
            $webPath = trim($webPath);

            if (!$webPath) {
                throw new Exception('Web path cannot be empty');
            }

            if ($webPath === 'sites') {
                throw new Exception("Web path 'sites' is prohibited");
            }

            if (array_key_exists($webPath, Config::getInstance()->get('sites') ?? [])) {
                throw new Exception('Site "' . $webPath . '" already exists.');
            }

            return $webPath;
        });
        return $question;
    }

    private function getLocalPathQuestion(): Question
    {
        $question = new Question('Local path');
        $question->setValidator(function (string $localPath): string {
            $localPath = trim($localPath);
            $localPath = Path::canonicalize($localPath);

            if (!file_exists($localPath)) {
                throw new Exception('Local path does not exist');
            }

            if (!is_readable($localPath)) {
                throw new Exception('Local path is not readable.');
            }

            return $localPath;

        });
        $question->setAutocompleterCallback(function (string $input): array {
            // Strip any characters from the last slash to the end of the string
            // to keep only the last directory and generate suggestions for it
            $inputPath = preg_replace('%(/|^)[^/]*$%', '$1', $input);
            $inputPath = '' === $inputPath ? getcwd() : $inputPath;

            if (
                !file_exists(Path::canonicalize($inputPath))
                || !is_readable(Path::canonicalize($inputPath))
            ) {
                return [];
            }

            // CAUTION - this example code allows unrestricted access to the
            // entire filesystem. In real applications, restrict the directories
            // where files and dirs can be found
            $foundFilesAndDirs = @scandir(Path::canonicalize($inputPath)) ?: [];

            return array_map(function (string $dirOrFile) use ($inputPath): string {
                return $inputPath . $dirOrFile;
            }, $foundFilesAndDirs);
        });
        return $question;
    }

    public function addSiteToConfig(mixed $webPath, string $source, ?string $mount): void
    {
        $config = Config::getInstance();
        $sites = $config->get('sites') ?? [];
        $sites[$webPath] = array_filter(compact('source', 'mount'));
        ksort($sites);
        $config->set('sites', $sites);

        $config->store();
    }
}